#ifndef _SMTLIB_SYNTAX_HH
#define _SMTLIB_SYNTAX_HH

#include <ostream>
#include <vector>
#include <string>
#include <set>
#include <gmpxx.h>
#include "formulaCompiler.hh"
#include "Solver.h"

class And;
class Or;

typedef std::set<std::string> identifierSet;

class Formula {
 public:
  void printIdentifier(std::ostream& o);

  virtual void print(std::ostream& o) const = 0;

  virtual ~Formula() = 0;

  virtual int size() const {
    return 1;
  }

  virtual void addToConjunction(And& conjunction);
  virtual void addToDisjunction(Or& conjunction);

  virtual void collectPropositionalVariables(identifierSet& s) const =0;  virtual void collectNumericVariables(identifierSet& s) const =0;
  virtual Lit compile(FormulaCompiler& compiler, Environment &env) const =0;
  virtual void printMathematica(std::ostream& out) const =0;
};

inline std::ostream& operator<<(std::ostream& o, const Formula& f) {
  f.print(o);
  return o;
}

class FLet : public Formula {
  std::string boundVariable;
  Formula *binding, *subFormula;

public:
  FLet(std::string v, Formula *b, Formula *f) :
    boundVariable(v), binding(b), subFormula(f) {
  }

  ~FLet() {
    delete binding;
    delete subFormula;
  }

  int size() const;  
  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  void print(std::ostream&) const;

  Lit compile(FormulaCompiler& compiler, Environment& env) const;
  void printMathematica(std::ostream& out) const;
};

class ManySubFormulas : public Formula {
protected:
  std::vector <Formula*> subFormulas;

public:
  typedef std::vector <Formula*>::const_iterator const_iterator;
  typedef std::vector <Formula*>::iterator iterator;

  iterator begin() {
    return subFormulas.begin();
  }
  iterator end() {
    return subFormulas.end();
  }

  const_iterator begin() const {
    return subFormulas.begin();
  }
  const_iterator end() const {
    return subFormulas.end();
  }

  ManySubFormulas() : subFormulas() {
  }

  ManySubFormulas(Formula* a, Formula* b) : subFormulas() {
    addFormula(a);
    addFormula(b);
  }

  ~ManySubFormulas();

  void addFormulaWithoutLooking(Formula *f) {
    subFormulas.push_back(f);
  }

  virtual void addFormula(Formula *f) {
    addFormulaWithoutLooking(f);
  }

  void printSubFormulas(std::ostream& o) const;

  int size() const;  
  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  void printSubtermsMathematica(std::ostream& out) const;
};

class And : public ManySubFormulas {
public:
  void print(std::ostream& o) const;

public:
  And() : ManySubFormulas() {
  }

  And(Formula *a, Formula *b) : ManySubFormulas(a, b) {
  }

  Lit compile(FormulaCompiler& compiler, Environment& env) const;

#ifdef COLLAPSE_COMBINATORS
  void addToConjunction(And& conjunction);
#endif

  void addFormula(Formula *f) {
    f->addToConjunction(*this);
  }

  void printMathematica(std::ostream& out) const;
};


class Iff : public ManySubFormulas {
public:
  void print(std::ostream& o) const;

public:
  Iff() : ManySubFormulas() {
  }

  Iff(Formula *a, Formula *b) : ManySubFormulas(a, b) {
  }

  Lit compile(FormulaCompiler& compiler, Environment& env) const;
  void printMathematica(std::ostream& out) const;
};

class Or : public ManySubFormulas {
public:
  void print(std::ostream& o) const;

public:
  Or() : ManySubFormulas() {
  }

  Or(Formula *a, Formula *b) : ManySubFormulas(a, b) {
  }

#ifdef COLLAPSE_COMBINATORS
  void addToDisjunction(Or& disjunction);
#endif

  Lit compile(FormulaCompiler& compiler, Environment& env) const;
  void printMathematica(std::ostream& out) const;
};

class PropositionalVariable : public Formula {
  std::string name;

public:
  void print(std::ostream& o) const;

  PropositionalVariable(const std::string& s) {
    name = s;
  }

  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;

  Lit compile(FormulaCompiler& compiler, Environment& env) const;
  void printMathematica(std::ostream& out) const;
};

class Not : public Formula {
  Formula *f;

public:
  void print(std::ostream& o) const;

  Not(Formula *p) : f(p) {
  }

  ~Not() {
    delete f;
  }

  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;

  Lit compile(FormulaCompiler& compiler, Environment& env) const;
  void printMathematica(std::ostream& out) const;
};

extern integer mpz_one, mpz_zero;

class Expression {
public:
  virtual void print(std::ostream& o) const = 0;

  virtual ~Expression() = 0;

  virtual int size() const {
    return 1;
  }

  virtual void collectPropositionalVariables(identifierSet& s) const =0;  virtual void collectNumericVariables(identifierSet& s) const =0;
  virtual CompiledExpression compile(FormulaCompiler& compiler, Environment &env) const =0;
  virtual void printMathematica(std::ostream& out) const =0;
};

class ManyExpression {
  std::vector<Expression*> expressions;

public:
  typedef std::vector<Expression*>::const_iterator const_iterator;
  
  const_iterator begin() const {
    return expressions.begin();
  }

  const_iterator end() const {
    return expressions.end();
  }

  void addExpression(Expression *e) {
    expressions.push_back(e);
  }

  ManyExpression() {
  }

  ManyExpression(Expression *e1, Expression *e2) {
    addExpression(e1);
    addExpression(e2);
  }

  ~ManyExpression();

  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  void printSubtermsMathematica(std::ostream& out) const;
};

class ManyExpressionExpression : public Expression, public ManyExpression {
  std::vector<Expression*> expressions;

public:
  ManyExpressionExpression() {
  }

  ManyExpressionExpression(Expression *e1, Expression *e2) :
    ManyExpression(e1, e2) {
  }

  void printCompoundExpression(std::ostream& o, std::string prefix) const;

  void collectPropositionalVariables(identifierSet& s) const {
    ManyExpression::collectPropositionalVariables(s);
  }

  void collectNumericVariables(identifierSet& s) const {
    ManyExpression::collectNumericVariables(s);
  }
};

class UnaryNeg : public Expression {
  Expression *f;

public:
  UnaryNeg(Expression *g) : f(g) {
  }

  ~UnaryNeg() {
    delete f;
  }

  void print(std::ostream& o) const;
  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  CompiledExpression compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class NumericVariable : public Expression {
  std::string name;

public:
  NumericVariable(const std::string& s) : name(s) {
  }  

  ~NumericVariable() {
  }

  void print(std::ostream& o) const;
  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  CompiledExpression compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class IntegerConstant : public Expression {
  integer value;

public:
  IntegerConstant(const integer& v) : value(v) {
  }

  ~IntegerConstant() {
  }

  void print(std::ostream& o) const;
  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  CompiledExpression compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

inline std::ostream& operator<<(std::ostream& o, const Expression& l) {
  l.print(o);
  return o;
}

class Sum : public ManyExpressionExpression {
public:
  Sum() {
  }

  Sum(Expression *e1, Expression *e2) : ManyExpressionExpression(e1, e2) {
  }

  void print(std::ostream& o) const;  
  CompiledExpression compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class Product : public ManyExpressionExpression {
public:
  Product() {
  }

  Product(Expression *e1, Expression *e2) : ManyExpressionExpression(e1, e2) {
  }

  void print(std::ostream& o) const;  
  CompiledExpression compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class Quotient : public Expression {
  Expression *e1, *e2;

public:
  ~Quotient() {
    delete e1;
    delete e2;
  }

  Quotient(Expression *f1, Expression *f2) : e1(f1), e2(f2) {
  }

  void print(std::ostream& o) const;  
  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  CompiledExpression compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class ManyExpressionFormula : public Formula, public ManyExpression {
public:
  ManyExpressionFormula() {
  }

  ManyExpressionFormula(Expression *a, Expression *b) : ManyExpression(a, b) {
  }

  int size() const {
    return 1;
  }

  void printListPredicate(std::ostream& o, std::string prefix) const;
  void collectPropositionalVariables(identifierSet& s) const {
    ManyExpression::collectPropositionalVariables(s);
  }
  void collectNumericVariables(identifierSet& s) const {
    ManyExpression::collectNumericVariables(s);
  }
};

class Equality : public ManyExpressionFormula {
public:
  Equality() {
  }

  Equality(Expression *a, Expression *b) : ManyExpressionFormula(a, b) {
  }

  void print(std::ostream& o) const;
  Lit compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class PairwiseDistinct : public ManyExpressionFormula {
public:
  PairwiseDistinct() {
  }

  PairwiseDistinct(Expression *a, Expression *b) : ManyExpressionFormula(a, b) {
  }

  void print(std::ostream& o) const;
  Lit compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class Comparison : public Formula {
  Expression *e1, *e2;

public:  
  enum ComparisonType {
    LESS_EQ, LESS, GREATER_EQ, GREATER };

private:
  enum ComparisonType comparisonKind;

public:
  ~Comparison() {
    delete e1;
    delete e2;
  }

  void print(std::ostream& o) const;

  ComparisonType readComparisonKind() {
    return comparisonKind;
  }

  Comparison(Expression *a1, ComparisonType cmp, Expression *a2)
  : e1(a1), e2(a2), comparisonKind(cmp) {
  }

  void collectNumericVariables(identifierSet& s) const;
  void collectPropositionalVariables(identifierSet& s) const;
  Lit compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

std::ostream& operator<<(std::ostream& o, Comparison::ComparisonType cmp);

class ExprLet : public Formula {
  std::string boundVariable;
  Expression *binding;
  Formula *subFormula;

public:
  ExprLet(std::string v, Expression *e, Formula *f) :
    boundVariable(v), binding(e), subFormula(f) {
  }

  ~ExprLet() {
    delete binding;
    delete subFormula;
  }

  int size() const;
  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  void print(std::ostream&) const;
  Lit compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class FormulaIfThenElse : public Formula {
  Formula *condition;
  Formula *e1, *e2;

public:
  FormulaIfThenElse(Formula *c, Formula *x1, Formula *x2) :
    condition(c), e1(x1), e2(x2) {
  }

  int size() const;
  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  void print(std::ostream&) const;

  ~FormulaIfThenElse();

  Lit compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};

class ExpressionIfThenElse : public Expression {
  Formula *condition;
  Expression *e1, *e2;

  std::string variableName() const;

public:
  ExpressionIfThenElse(Formula *c, Expression *x1, Expression *x2) :
    condition(c), e1(x1), e2(x2) {
  }

  void collectPropositionalVariables(identifierSet& s) const;
  void collectNumericVariables(identifierSet& s) const;
  void print(std::ostream&) const;

  ~ExpressionIfThenElse();

  CompiledExpression compile(FormulaCompiler& compiler, Environment &env) const;
  void printMathematica(std::ostream& out) const;
};
#endif
